from enum import auto, Enum

class HealthStatus(Enum):
    GREEN = auto()
    YELLOW = auto()
    ORANGE = auto()
    RED = auto()
    GRAY = auto()


class SeniorityLevel(Enum):
    JUNIOR = auto()
    MIDDLE = auto()
    SENIOR = auto()



from abc import ABC, abstractmethod
from typing import Dict, Callable, Optional, TypeVar

from sd_metrics_lib.utils.time import Duration, TimeUnit

from sd_metrics_lib.utils.attributes import get_attribute_by_path

T = TypeVar('T')


class WorklogExtractor(ABC):

    @abstractmethod
    def get_work_time_per_user(self, task) -> Dict[str, 'Duration']:
        pass


class TaskTotalSpentTimeExtractor(ABC):

    @abstractmethod
    def get_total_spent_time(self, task) -> 'Duration':
        pass


class ChainedWorklogExtractor(WorklogExtractor):

    def __init__(self, worklog_extractor_list: list[WorklogExtractor]) -> None:
        self.worklog_extractor_list = worklog_extractor_list

    def get_work_time_per_user(self, task):
        for worklog_extractor in self.worklog_extractor_list:
            work_time = worklog_extractor.get_work_time_per_user(task)
            if work_time is not None and len(work_time.keys()) != 0:
                return work_time
        return {}


class FunctionWorklogExtractor(WorklogExtractor):

    def __init__(self, func: Callable[[T], Optional[Dict[str, Duration]]]):
        self.func = func

    def get_work_time_per_user(self, task: T) -> Dict[str, Duration]:
        result = self.func(task)
        try:
            return {str(k): v for k, v in (result or {}).items()}
        except Exception:
            return {}


class FunctionTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, func: Callable[[T], Optional[Duration]]):
        self.func = func

    def get_total_spent_time(self, task: T) -> Duration:
        result = self.func(task)
        try:
            return result if isinstance(result, Duration) else Duration.zero()
        except Exception:
            return Duration.zero()


class AttributePathWorklogExtractor(WorklogExtractor):

    def __init__(self, attr_path: str):
        self._path = attr_path

    def get_work_time_per_user(self, task) -> Dict[str, Duration]:
        value = get_attribute_by_path(task, self._path, {})
        if isinstance(value, dict):
            try:
                return {str(k): v for k, v in value.items()}
            except Exception:
                return {}
        return {}


class AttributePathTotalSpentTimeExtractor(TaskTotalSpentTimeExtractor):

    def __init__(self, attr_path: str, default_seconds: float = 0):
        self._path = attr_path
        self._default = Duration.of(default_seconds, TimeUnit.SECOND)

    def get_total_spent_time(self, task) -> Duration:
        value = get_attribute_by_path(task, self._path, self._default)
        try:
            return value if isinstance(value, Duration) else self._default
        except Exception:
            return self._default



import unittest
from sd_metrics_lib.utils.time import Duration, TimeUnit


class TestDurationNumericArithmetic(unittest.TestCase):
    def test_add_number_on_right_adds_in_same_unit(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertEqual(duration_two_hours + 3, Duration.of(5, TimeUnit.HOUR))

    def test_add_number_on_left_adds_in_same_unit(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertEqual(3 + duration_two_hours, Duration.of(5, TimeUnit.HOUR))

    def test_sub_number_on_right_subtracts_in_same_unit(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertEqual(duration_two_hours - 0.5, Duration.of(1.5, TimeUnit.HOUR))

    def test_sub_number_on_left_subtracts_duration_from_number(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertEqual(2.5 - duration_two_hours, Duration.of(0.5, TimeUnit.HOUR))

    def test_iadd_with_number_returns_new_instance_sum(self):
        duration_one_day = Duration.of(1, TimeUnit.DAY)
        duration_one_day = duration_one_day.__iadd__(2)
        self.assertEqual(duration_one_day, Duration.of(3, TimeUnit.DAY))

    def test_isub_with_number_returns_new_instance_difference(self):
        duration_three_days = Duration.of(3, TimeUnit.DAY)
        duration_three_days = duration_three_days.__isub__(1.5)
        self.assertEqual(duration_three_days, Duration.of(1.5, TimeUnit.DAY))

    def test_mul_right_by_number_multiplies_value(self):
        duration_four_hours = Duration.of(4, TimeUnit.HOUR)
        self.assertEqual(duration_four_hours * 2, Duration.of(8, TimeUnit.HOUR))

    def test_mul_left_by_number_multiplies_value(self):
        duration_four_hours = Duration.of(4, TimeUnit.HOUR)
        self.assertEqual(2 * duration_four_hours, Duration.of(8, TimeUnit.HOUR))

    def test_div_right_by_number_divides_value(self):
        duration_four_hours = Duration.of(4, TimeUnit.HOUR)
        self.assertEqual((duration_four_hours / 4), Duration.of(1, TimeUnit.HOUR))

    def test_division_duration_by_duration_equal_returns_ratio_one(self):
        duration_forty_eight_hours = Duration.of(48, TimeUnit.HOUR)
        duration_two_days = Duration.of(2, TimeUnit.DAY)
        self.assertAlmostEqual(duration_forty_eight_hours / duration_two_days, 1.0, places=12)

    def test_division_duration_by_duration_different_values_returns_ratio_two(self):
        duration_forty_eight_hours = Duration.of(48, TimeUnit.HOUR)
        duration_one_day = Duration.of(1, TimeUnit.DAY)
        self.assertAlmostEqual(duration_forty_eight_hours / duration_one_day, 2.0, places=12)

    def test_number_divided_by_duration_raises_type_error(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        with self.assertRaises(TypeError):
            5 / duration_two_hours  # type: ignore[operator]

    def test_add_duration_with_mixed_units_uses_default_policy_result_in_days(self):
        duration_one_day = Duration.of(1, TimeUnit.DAY)
        duration_four_hours = Duration.of(4, TimeUnit.HOUR)
        summed_in_days = duration_one_day + duration_four_hours
        self.assertEqual(summed_in_days, Duration.of(1 + 4 / 24, TimeUnit.DAY))

    def test_sub_duration_with_mixed_units_uses_default_policy_result_in_days(self):
        duration_one_day = Duration.of(1, TimeUnit.DAY)
        duration_four_hours = Duration.of(4, TimeUnit.HOUR)
        difference_in_days = duration_one_day - duration_four_hours
        self.assertEqual(difference_in_days, Duration.of(1 - 4 / 24, TimeUnit.DAY))


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass
from datetime import datetime

from enum import Enum, auto
from typing import Iterable, ClassVar, SupportsFloat


class TimeUnit(Enum):
    SECOND = auto()
    HOUR = auto()
    DAY = auto()
    WEEK = auto()
    MONTH = auto()


SECONDS_IN_HOUR = 3600
WORKING_HOURS_PER_DAY = 8
WORKING_DAYS_PER_WEEK = 5
WORKING_WEEKS_IN_MONTH = 4

# Python's date.weekday(): Monday=0
WEEKDAY_FRIDAY = 4


@dataclass(frozen=True, slots=True)
class TimePolicy:
    hours_per_day: float
    days_per_week: float
    days_per_month: float

    ALL_HOURS: ClassVar["TimePolicy"]  # 24/7 wall-clock (aka civil)
    BUSINESS_HOURS: ClassVar["TimePolicy"]  # working capacity (e.g., 8h/day, 5d/week)

    def factor_to_day(self, unit: TimeUnit) -> float:
        if unit == TimeUnit.SECOND:
            return 1.0 / (SECONDS_IN_HOUR * self.hours_per_day)
        if unit == TimeUnit.HOUR:
            return 1.0 / self.hours_per_day
        if unit == TimeUnit.DAY:
            return 1.0
        if unit == TimeUnit.WEEK:
            return self.days_per_week
        if unit == TimeUnit.MONTH:
            return self.days_per_month
        return 1.0

    def factor_from_day(self, unit: TimeUnit) -> float:
        if unit == TimeUnit.SECOND:
            return SECONDS_IN_HOUR * self.hours_per_day
        if unit == TimeUnit.HOUR:
            return self.hours_per_day
        if unit == TimeUnit.DAY:
            return 1.0
        if unit == TimeUnit.WEEK:
            return 1.0 / self.days_per_week
        if unit == TimeUnit.MONTH:
            return 1.0 / self.days_per_month
        return 1.0

    def convert(self, value: float, from_unit: TimeUnit, to_unit: TimeUnit) -> float:
        if from_unit == to_unit:
            return float(value)
        return float(value) * self.factor_to_day(from_unit) * self.factor_from_day(to_unit)


TimePolicy.ALL_HOURS = TimePolicy(
    hours_per_day=24,
    days_per_week=7,
    days_per_month=30,
)

TimePolicy.BUSINESS_HOURS = TimePolicy(
    hours_per_day=WORKING_HOURS_PER_DAY,
    days_per_week=WORKING_DAYS_PER_WEEK,
    days_per_month=WORKING_DAYS_PER_WEEK * WORKING_WEEKS_IN_MONTH,
)


@dataclass(frozen=True, slots=True)
class Duration:
    time_delta: float
    time_unit: TimeUnit

    @classmethod
    def zero(cls, unit: TimeUnit = TimeUnit.SECOND) -> "Duration":
        return cls(0.0, unit)

    @classmethod
    def of(cls, time_delta: SupportsFloat, time_unit: TimeUnit) -> "Duration":
        return cls(float(time_delta), time_unit)

    @classmethod
    def difference(cls,
                   start_value: SupportsFloat,
                   end_value: SupportsFloat,
                   time_unit: TimeUnit) -> "Duration":
        return cls(float(end_value) - float(start_value), time_unit)

    @classmethod
    def datetime_difference(
            cls,
            start: datetime,
            end: datetime,
            time_unit: TimeUnit,
            time_policy: TimePolicy | None = None,
    ) -> "Duration":
        policy = time_policy or TimePolicy.ALL_HOURS
        delta_seconds = float((end - start).total_seconds())
        value_in_unit = policy.convert(delta_seconds, TimeUnit.SECOND, time_unit)
        return cls(value_in_unit, time_unit)

    def to_seconds(self, time_policy: TimePolicy | None = None) -> float:
        policy = time_policy or TimePolicy.ALL_HOURS
        return policy.convert(self.time_delta, self.time_unit, TimeUnit.SECOND)

    def convert(self, target_unit: TimeUnit, time_policy: TimePolicy | None = None) -> "Duration":
        policy = time_policy or TimePolicy.ALL_HOURS
        new_value = policy.convert(self.time_delta, self.time_unit, target_unit)
        return Duration(new_value, target_unit)

    def is_zero(self, eps: float = 0.0, time_policy: TimePolicy | None = None) -> bool:
        if eps <= 0:
            return self.time_delta == 0
        return abs(self.to_seconds(time_policy)) <= eps

    def add(self, other: "Duration", policy: TimePolicy | None = None, unit: TimeUnit | None = None) -> "Duration":
        unit_used = unit or self.time_unit
        policy_used = policy or TimePolicy.ALL_HOURS
        self_value_in_unit = policy_used.convert(self.time_delta, self.time_unit, unit_used)
        other_value_in_unit = policy_used.convert(other.time_delta, other.time_unit, unit_used)
        return Duration.of(self_value_in_unit + other_value_in_unit, unit_used)

    def sub(self, other: "Duration", policy: TimePolicy | None = None, unit: TimeUnit | None = None) -> "Duration":
        unit_used = unit or self.time_unit
        policy_used = policy or TimePolicy.ALL_HOURS
        self_value_in_unit = policy_used.convert(self.time_delta, self.time_unit, unit_used)
        other_value_in_unit = policy_used.convert(other.time_delta, other.time_unit, unit_used)
        return Duration.of(self_value_in_unit - other_value_in_unit, unit_used)

    @staticmethod
    def _coerce_value_in_unit(other: object, target_unit: TimeUnit,
                              time_policy: TimePolicy | None = None) -> float | None:
        if isinstance(other, Duration):
            effective_policy = time_policy or TimePolicy.ALL_HOURS
            return effective_policy.convert(other.time_delta, other.time_unit, target_unit)
        if isinstance(other, (int, float)):
            return float(other)
        return None

    def __add__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented
        return Duration.of(self.time_delta + coerced_value, self.time_unit)

    def __radd__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented
        return Duration.of(coerced_value + self.time_delta, self.time_unit)

    def __sub__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented
        return Duration.of(self.time_delta - coerced_value, self.time_unit)

    def __rsub__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented
        return Duration.of(coerced_value - self.time_delta, self.time_unit)

    def __iadd__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented  # type: ignore[return-value]
        return Duration.of(self.time_delta + coerced_value, self.time_unit)

    def __isub__(self, other: object) -> "Duration":
        coerced_value = self._coerce_value_in_unit(other, self.time_unit)
        if coerced_value is None:
            return NotImplemented  # type: ignore[return-value]
        return Duration.of(self.time_delta - coerced_value, self.time_unit)

    def __mul__(self, multiplier: SupportsFloat) -> "Duration":
        return Duration.of(float(self.time_delta) * float(multiplier), self.time_unit)

    def __rmul__(self, multiplier: SupportsFloat) -> "Duration":
        return Duration.of(float(multiplier) * float(self.time_delta), self.time_unit)

    def __truediv__(self, other: object):
        if isinstance(other, (int, float)):
            return Duration.of(float(self.time_delta) / float(other), self.time_unit)
        if isinstance(other, Duration):
            denominator_value = self._coerce_value_in_unit(other, self.time_unit)
            if denominator_value is None:
                return NotImplemented
            return float(self.time_delta) / float(denominator_value)
        return NotImplemented

    def _cmp_seconds(self, other: object) -> float | None:
        if isinstance(other, Duration):
            self_seconds = self.to_seconds(TimePolicy.ALL_HOURS)
            other_seconds = other.to_seconds(TimePolicy.ALL_HOURS)
            return self_seconds - other_seconds
        if isinstance(other, (int, float)):
            self_value = self.time_delta
            other_value = float(other)
            return self_value - other_value
        return None

    def __eq__(self, other: object) -> bool:
        comparison_diff = self._cmp_seconds(other)
        if comparison_diff is None:
            return NotImplemented
        return comparison_diff == 0

    def __lt__(self, other: object) -> bool:
        comparison_diff = self._cmp_seconds(other)
        if comparison_diff is None:
            return NotImplemented
        return comparison_diff < 0

    def __le__(self, other: object) -> bool:
        comparison_diff = self._cmp_seconds(other)
        if comparison_diff is None:
            return NotImplemented
        return comparison_diff <= 0

    def __gt__(self, other: object) -> bool:
        comparison_diff = self._cmp_seconds(other)
        if comparison_diff is None:
            return NotImplemented
        return comparison_diff > 0

    def __ge__(self, other: object) -> bool:
        comparison_diff = self._cmp_seconds(other)
        if comparison_diff is None:
            return NotImplemented
        return comparison_diff >= 0

    def __bool__(self) -> bool:
        return self.time_delta != 0

    @staticmethod
    def sum(durations: "Iterable[Duration]", policy: TimePolicy | None = None,
            unit: TimeUnit = TimeUnit.SECOND) -> "Duration":
        policy_used = policy or TimePolicy.ALL_HOURS
        total = 0.0
        if durations:
            for duration in durations:
                total += policy_used.convert(duration.time_delta, duration.time_unit, unit)
        return Duration.of(total, unit)



import datetime
from abc import ABC, abstractmethod

from sd_metrics_lib.utils.time import WEEKDAY_FRIDAY, Duration, TimeUnit, TimePolicy


class WorkTimeExtractor(ABC):

    @abstractmethod
    def extract_time_from_period(self, start_time_period: datetime.date | datetime.datetime, end_time_period: datetime.date | datetime.datetime) -> Duration | None:
        pass


class SimpleWorkTimeExtractor(WorkTimeExtractor):

    def extract_time_from_period(self,
                                 start_time_period: datetime.date | datetime.datetime,
                                 end_time_period: datetime.date | datetime.datetime) -> Duration | None:
        if end_time_period <= start_time_period:
            return None

        # Compute civil elapsed time only in seconds; derive days when needed via conversion
        period_duration = Duration.datetime_difference(start_time_period, end_time_period, TimeUnit.SECOND)

        # Ignore short periods
        minimal_period_length_limit = Duration.of(0.25, TimeUnit.HOUR).convert(TimeUnit.SECOND)
        if period_duration < minimal_period_length_limit:
            return None

        # Limit multiday periods with week working days and work working hours in a day
        # In other words, a full week period is converted into 5 days with 8 hours in each day
        period_duration_in_days = period_duration.convert(TimeUnit.DAY)
        if period_duration_in_days.time_delta >= 1.0:
            work_days = self.__count_work_days(start_time_period, end_time_period)
            round_up_period_days = int(period_duration_in_days.time_delta) + 1
            return Duration.of(min(work_days, round_up_period_days), TimeUnit.DAY).convert(TimeUnit.SECOND, TimePolicy.BUSINESS_HOURS)

        # Limit 24-hour period with working hours in a single day
        one_business_day = Duration.of(1, TimeUnit.DAY).convert(TimeUnit.SECOND, TimePolicy.BUSINESS_HOURS)
        if period_duration.time_delta < one_business_day.time_delta:
            return period_duration
        return one_business_day

    @staticmethod
    def __count_work_days(start_date: datetime.date, end_date: datetime.date):
        # Move start forward to Monday if it falls on weekend
        if start_date.weekday() > WEEKDAY_FRIDAY:
            start_date = start_date + datetime.timedelta(days=7 - start_date.weekday())
        # Move end backward to Friday if it falls on weekend
        if end_date.weekday() > WEEKDAY_FRIDAY:
            end_date = end_date - datetime.timedelta(days=end_date.weekday() - WEEKDAY_FRIDAY)

        if start_date > end_date:
            return 0

        # Inclusive range
        total_days = (end_date - start_date).days + 1
        full_weeks = total_days // 7
        remainder = total_days % 7

        workdays = full_weeks * 5
        start_wd = start_date.weekday()
        for i in range(remainder):
            if (start_wd + i) % 7 <= WEEKDAY_FRIDAY:
                workdays += 1
        return workdays

SIMPLE_WORKTIME_EXTRACTOR = SimpleWorkTimeExtractor()

class BoundarySimpleWorkTimeExtractor(SimpleWorkTimeExtractor):

    def __init__(self,
                 start_time_boundary: datetime.date,
                 end_time_boundary: datetime.date) -> None:
        self.start_time_boundary = start_time_boundary
        self.end_time_boundary = end_time_boundary

    def extract_time_from_period(self, start_time_period: datetime.date | datetime.datetime, end_time_period: datetime.date | datetime.datetime) -> Duration | None:

        if self.start_time_boundary < start_time_period:
            new_start_period = start_time_period
        else:
            new_start_period = self.start_time_boundary

        if self.end_time_boundary > end_time_period:
            new_end_time_period = end_time_period
        else:
            new_end_time_period = self.end_time_boundary

        return super().extract_time_from_period(new_start_period, new_end_time_period)



import unittest
from datetime import datetime, timedelta

from sd_metrics_lib.utils.time import TimeUnit, TimePolicy, Duration, SECONDS_IN_HOUR, WORKING_HOURS_PER_DAY, \
    WORKING_DAYS_PER_WEEK, WORKING_WEEKS_IN_MONTH


class TestTimePolicy(unittest.TestCase):
    def test_all_hours_seconds_to_hour(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_seconds_to_hours = Duration.of(all_hours_policy.convert(SECONDS_IN_HOUR, TimeUnit.SECOND, TimeUnit.HOUR), TimeUnit.HOUR)
        expected_one_hour_duration = Duration.of(1.0, TimeUnit.HOUR)
        self.assertEqual(converted_seconds_to_hours, expected_one_hour_duration)

    def test_all_hours_seconds_to_day(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_seconds_to_days = Duration.of(all_hours_policy.convert(SECONDS_IN_HOUR * 24, TimeUnit.SECOND, TimeUnit.DAY), TimeUnit.DAY)
        expected_one_day_duration = Duration.of(1.0, TimeUnit.DAY)
        self.assertEqual(converted_seconds_to_days, expected_one_day_duration)

    def test_all_hours_day_to_seconds(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_one_day_to_seconds = Duration.of(all_hours_policy.convert(1, TimeUnit.DAY, TimeUnit.SECOND), TimeUnit.SECOND)
        expected_seconds_in_day = Duration.of(24 * SECONDS_IN_HOUR, TimeUnit.SECOND)
        self.assertEqual(converted_one_day_to_seconds, expected_seconds_in_day)

    def test_all_hours_hours_to_day(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_24h_to_days = Duration.of(all_hours_policy.convert(24, TimeUnit.HOUR, TimeUnit.DAY), TimeUnit.DAY)
        expected_one_day = Duration.of(1.0, TimeUnit.DAY)
        self.assertEqual(converted_24h_to_days, expected_one_day)

    def test_all_hours_week_to_day(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_week_to_days = Duration.of(all_hours_policy.convert(1, TimeUnit.WEEK, TimeUnit.DAY), TimeUnit.DAY)
        expected_seven_days = Duration.of(7, TimeUnit.DAY)
        self.assertEqual(converted_week_to_days, expected_seven_days)

    def test_all_hours_month_to_day(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_month_to_days = Duration.of(all_hours_policy.convert(1, TimeUnit.MONTH, TimeUnit.DAY), TimeUnit.DAY)
        expected_thirty_days = Duration.of(30, TimeUnit.DAY)
        self.assertEqual(converted_month_to_days, expected_thirty_days)

    def test_all_hours_identity(self):
        all_hours_policy = TimePolicy.ALL_HOURS
        converted_identity_hours = Duration.of(all_hours_policy.convert(3.14, TimeUnit.HOUR, TimeUnit.HOUR), TimeUnit.HOUR)
        expected_same_hours = Duration.of(3.14, TimeUnit.HOUR)
        self.assertEqual(converted_identity_hours, expected_same_hours)

    def test_business_hours_hour_to_day(self):
        business_hours_policy = TimePolicy.BUSINESS_HOURS
        converted_working_hours_per_day_to_day = Duration.of(business_hours_policy.convert(WORKING_HOURS_PER_DAY, TimeUnit.HOUR, TimeUnit.DAY), TimeUnit.DAY)
        expected_one_business_day = Duration.of(1.0, TimeUnit.DAY)
        self.assertEqual(converted_working_hours_per_day_to_day, expected_one_business_day)

    def test_business_hours_day_to_seconds(self):
        business_hours_policy = TimePolicy.BUSINESS_HOURS
        converted_business_day_to_seconds = Duration.of(business_hours_policy.convert(1, TimeUnit.DAY, TimeUnit.SECOND), TimeUnit.SECOND)
        expected_seconds_in_business_day = Duration.of(WORKING_HOURS_PER_DAY * SECONDS_IN_HOUR, TimeUnit.SECOND)
        self.assertEqual(converted_business_day_to_seconds, expected_seconds_in_business_day)

    def test_business_hours_week_to_day(self):
        business_hours_policy = TimePolicy.BUSINESS_HOURS
        converted_business_week_to_days = Duration.of(business_hours_policy.convert(1, TimeUnit.WEEK, TimeUnit.DAY), TimeUnit.DAY)
        expected_working_days_per_week = Duration.of(WORKING_DAYS_PER_WEEK, TimeUnit.DAY)
        self.assertEqual(converted_business_week_to_days, expected_working_days_per_week)

    def test_business_hours_month_to_day(self):
        business_hours_policy = TimePolicy.BUSINESS_HOURS
        converted_business_month_to_days = Duration.of(business_hours_policy.convert(1, TimeUnit.MONTH, TimeUnit.DAY), TimeUnit.DAY)
        expected_working_days_per_month = Duration.of(WORKING_DAYS_PER_WEEK * WORKING_WEEKS_IN_MONTH, TimeUnit.DAY)
        self.assertEqual(converted_business_month_to_days, expected_working_days_per_month)

    def test_business_hours_round_trip_via_day(self):
        business_hours_policy = TimePolicy.BUSINESS_HOURS
        value = 13.7
        for unit in list(TimeUnit):
            with self.subTest(unit=unit):
                roundtrip_converted = Duration.of(
                    business_hours_policy.convert(
                        business_hours_policy.convert(value, unit, TimeUnit.DAY),
                        TimeUnit.DAY,
                        unit
                    ),
                    unit
                )
                original_value_duration = Duration.of(value, unit)
                self.assertTrue(roundtrip_converted.sub(original_value_duration).is_zero(eps=1e-6))


class TestDuration(unittest.TestCase):
    def test_zero_returns_false_bool(self):
        z = Duration.zero(TimeUnit.HOUR)
        assert not z

    def test_zero_keeps_unit_and_value(self):
        z = Duration.zero(TimeUnit.HOUR)
        assert z.time_delta == 0.0 and z.time_unit == TimeUnit.HOUR

    def test_of_constructs(self):
        d = Duration.of(2, TimeUnit.DAY)
        assert d and d.time_delta == 2.0 and d.time_unit == TimeUnit.DAY

    def test_difference_numeric(self):
        d = Duration.difference(10, 25.5, TimeUnit.HOUR)
        assert d == Duration.of(15.5, TimeUnit.HOUR)

    def test_datetime_difference_hours(self):
        start = datetime(2025, 1, 1, 0, 0, 0)
        end = start + timedelta(days=2, hours=6)
        d_hours = Duration.datetime_difference(start, end, TimeUnit.HOUR)
        self.assertEqual(d_hours, Duration.of(54.0, TimeUnit.HOUR))

    def test_datetime_difference_days(self):
        start = datetime(2025, 1, 1, 0, 0, 0)
        end = start + timedelta(days=2, hours=6)
        d_days = Duration.datetime_difference(start, end, TimeUnit.DAY)
        self.assertEqual(d_days, Duration.of(2.25, TimeUnit.DAY))

    def test_to_seconds_all_hours(self):
        d = Duration.of(2, TimeUnit.DAY)
        converted_two_days_to_seconds = Duration.of(d.to_seconds(TimePolicy.ALL_HOURS), TimeUnit.SECOND)
        expected_seconds = Duration.of(2 * 24 * SECONDS_IN_HOUR, TimeUnit.SECOND)
        self.assertEqual(converted_two_days_to_seconds, expected_seconds)

    def test_convert_business_day_to_hours(self):
        d = Duration.of(2, TimeUnit.DAY)
        converted_business_days_to_hours = d.convert(TimeUnit.HOUR, TimePolicy.BUSINESS_HOURS)
        expected_business_hours = Duration.of(2 * WORKING_HOURS_PER_DAY, TimeUnit.HOUR)
        self.assertEqual(converted_business_days_to_hours, expected_business_hours)

    def test_convert_business_hours_to_days(self):
        h = Duration.of(16, TimeUnit.HOUR)
        converted_hours_to_business_days = h.convert(TimeUnit.DAY, TimePolicy.BUSINESS_HOURS)
        expected_two_business_days = Duration.of(2.0, TimeUnit.DAY)
        self.assertEqual(converted_hours_to_business_days, expected_two_business_days)

    def test_is_zero_exact_zero(self):
        assert Duration.zero(TimeUnit.SECOND).is_zero()

    def test_is_zero_with_epsilon_seconds(self):
        tiny = Duration.of(0.5, TimeUnit.SECOND)
        assert tiny.is_zero(eps=1.0)

    def test_is_zero_with_epsilon_hour_equivalent(self):
        tiny_h = Duration.of(1 / SECONDS_IN_HOUR * 0.5, TimeUnit.HOUR)
        assert tiny_h.is_zero(eps=1.0)

    def test_is_zero_eps_zero_means_exact(self):
        assert not Duration.of(1e-12, TimeUnit.SECOND).is_zero(eps=0.0)

    def test_add_default_policy(self):
        a = Duration.of(1, TimeUnit.DAY)
        b = Duration.of(4, TimeUnit.HOUR)
        c = a + b
        assert c == Duration.of(1 + 4 / 24, TimeUnit.DAY)

    def test_sub_default_policy(self):
        a = Duration.of(1, TimeUnit.DAY)
        b = Duration.of(4, TimeUnit.HOUR)
        d = a - b
        assert d == Duration.of(1 - 4 / 24, TimeUnit.DAY)

    def test_add_with_business_hours_target_hour(self):
        a = Duration.of(1, TimeUnit.DAY)
        b = Duration.of(4, TimeUnit.HOUR)
        c2 = a.add(b, policy=TimePolicy.BUSINESS_HOURS, unit=TimeUnit.HOUR)
        assert c2 == Duration.of(WORKING_HOURS_PER_DAY + 4, TimeUnit.HOUR)

    def test_scalar_operations_multiplication_division(self):
        b = Duration.of(4, TimeUnit.HOUR)
        e = (b * 2.5) / 5
        assert e == Duration.of(2.0, TimeUnit.HOUR)

    def test_equality_cross_units(self):
        two_days = Duration.of(2, TimeUnit.DAY)
        forty_eight_hours = Duration.of(48, TimeUnit.HOUR)
        assert two_days == forty_eight_hours or (
                    not (two_days < forty_eight_hours) and not (two_days > forty_eight_hours))

    def test_le_ge_cross_units(self):
        two_days = Duration.of(2, TimeUnit.DAY)
        forty_eight_hours = Duration.of(48, TimeUnit.HOUR)
        assert two_days <= forty_eight_hours and two_days >= forty_eight_hours

    def test_week_greater_than_six_days(self):
        one_week = Duration.of(1, TimeUnit.WEEK)
        six_days = Duration.of(6, TimeUnit.DAY)
        assert one_week > six_days

    def test_sum_mixed_units_all_hours_hours(self):
        items = [
            Duration.of(1, TimeUnit.DAY),
            Duration.of(4, TimeUnit.HOUR),
            Duration.of(30, TimeUnit.MINUTE) if hasattr(TimeUnit, 'MINUTE') else Duration.of(0.5, TimeUnit.HOUR),
        ]
        total = Duration.sum(items, policy=TimePolicy.ALL_HOURS, unit=TimeUnit.HOUR)
        assert total == Duration.of(28.5, TimeUnit.HOUR)

    def test_sum_mixed_units_business_days(self):
        items = [
            Duration.of(1, TimeUnit.DAY),
            Duration.of(4, TimeUnit.HOUR),
            Duration.of(30, TimeUnit.MINUTE) if hasattr(TimeUnit, 'MINUTE') else Duration.of(0.5, TimeUnit.HOUR),
        ]
        total_b = Duration.sum(items, policy=TimePolicy.BUSINESS_HOURS, unit=TimeUnit.DAY)
        assert total_b == Duration.of((WORKING_HOURS_PER_DAY + 4 + 0.5) / WORKING_HOURS_PER_DAY, TimeUnit.DAY)


if __name__ == '__main__':
    unittest.main()



import unittest
from sd_metrics_lib.utils.time import Duration, TimeUnit


class TestDurationNumericComparison(unittest.TestCase):
    def test_eq_number_in_same_unit_matches_value(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertEqual(duration_two_hours, 2)

    def test_ne_number_in_same_unit_is_false_when_equal(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertFalse(duration_two_hours != 2)

    def test_gt_number_in_same_unit_when_greater(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertTrue(duration_two_hours > 1.5)

    def test_ge_number_in_same_unit_when_equal(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertTrue(duration_two_hours >= 2)

    def test_lt_number_in_same_unit_when_less(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertTrue(duration_two_hours < 3)

    def test_le_number_in_same_unit_when_equal(self):
        duration_two_hours = Duration.of(2, TimeUnit.HOUR)
        self.assertTrue(duration_two_hours <= 2)

    def test_eq_number_uses_duration_unit_semantics(self):
        duration_twenty_four_hours = Duration.of(24, TimeUnit.HOUR)
        self.assertEqual(duration_twenty_four_hours, 24)

    def test_number_comparison_from_left_is_not_reflected_but_right_side_works(self):
        duration_three_hours = Duration.of(3, TimeUnit.HOUR)
        self.assertTrue(duration_three_hours > 2)


if __name__ == '__main__':
    unittest.main()
